import numpy as np
from tensorflow.keras.layers import Input, Dense, Lambda, Activation, BatchNormalization, Concatenate, Flatten, Reshape
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam
from tensorflow.keras import regularizers
import tensorflow.keras.backend as K
from sklearn.metrics import r2_score, mean_squared_error

### Sampling function and loss function implementation based upon:
### https://console.paperspace.com/ml-showcase/notebook/r0uxmfm6hvw0duf?file=VAE.ipynb

# sampling function for latent vectors

def sampling_function(args):

    # pass mean and log_variance
    
    mean, log_variance = args

    # epislon = random samples from Gaussian distribution in shape of mean

    epsilon = K.random_normal(shape=(K.shape(mean)))

    # z = latent distribution vector

    z = mean + K.exp(log_variance/2) * epsilon

    return z

# loss function - comprises weighted mean squared error and Kullback Leibler Divergence
# loss functions wrapped in an outer function to enable passing of mean and log variance for kl divergence

def loss_function(mean, log_variance):

    # reconstruction loss = mean squared error
    
    def reconstruction_loss(y_target, y_predict):
        
            rl = K.mean(K.square(y_target - y_predict), axis = 1)
            
            return rl
    
    # kl divergence = here is divergence from normal distribution
    # from Kingma and Welling 2013
    # in Kingma Welling 2013 the equation is for negative kl divergence, therefore negative added here

    def kl_divergence(mean, log_variance):
        
        kl_divergence = -0.5 * K.sum(1 + log_variance - K.square(mean) - K.exp(log_variance), axis=1)
        
        return kl_divergence

    # loss = weighting of reconstruction loss and kl divergence - weighted towards reconstruction here

    def loss(y_target, y_predict):
        
            rl = reconstruction_loss(y_target, y_predict)
            
            kl = kl_divergence(y_target, y_predict)

            global rl_weight
            global kl_weight
            
            loss = rl_weight * rl + kl_weight * kl
            
            return loss

    return loss

if __name__ == "__main__":

    # set audio/video variables

    fps = 60 # set camera frames per second
    segment_length = 0.5 # length of sample segment in seconds
    segment_frames = int(fps*segment_length) # set the number of video frames per segment
    aspect_x = 64 # aspect ratio for x
    aspect_y = 48 # aspect ratio for y

    # load audio and motiongrams

    audio = np.load('./Temp/audio_features.npy')
    mgs_x = np.load('./Temp/motion_features_x.npy')
    mgs_y = np.load('./Temp/motion_features_y.npy')

    # normalise to range 1, -1

    mgs_x = (mgs_x/128) -1
    mgs_y = (mgs_y/128) -1

    # set VAE variables

    input_1_size = audio.shape[1] # set audio input size
    input_2_size = mgs_x.shape[1], mgs_x.shape[2], mgs_x.shape[3] # set motiongram x input size
    input_3_size = mgs_y.shape[1], mgs_y.shape[2], mgs_y.shape[3] # set motiongram y input size
    latent_space_size = 512 # set latent space size
    epochs = 50 # set number of epochs to train
    batch_size = 256 # set batch size
    learning_rate = 0.0001 # set learning rate
    rl_weight = 16000 # set reconstruction loss weighting for loss function
    kl_weight = 0.001 # set Kullback-Leibler divergence weighting for loss function
    L2_regulariser = 1 # set the L2 regularisation value

    ### Encoder ###

    # This is structured as an audio encoder, mg x encoder, and mg y encoder that are concatenated together
    # and passed to a lambda layer. Each of these are created as separate models that are then put together as
    # an entire encoder model. This is to allow a modular structure in that these individual models can be used
    # to create a new inference model comprising individual elements of the training model.

    # For this instance the audio, mg x and mg y encoders simply pass the input. This is to reduce trainable parameters
    # and decrease inference time. Howeve, extra layers can be added if desired, however these need to then also be added
    # to the inference model used in the performance system. 

    ### Audio Encoder ###

    # Here the input is just passed through in order to keep the number of parameters and therefore the inference time low
    # On more powerful computers extra layers can be added here in order to increase the performance of the model

    audio_input = Input(shape=input_1_size, name='audio_input')

    audio_encoder = Model(audio_input, audio_input, name='audio_encoder')

    ### Motiongram x Encoder ###

    # Here the input is just passed through in order to keep the number of parameters and therefore the inference time low
    # On more powerful computers extra layers can be added here in order to increase the performance of the model

    mg_x_input = Input(shape=input_2_size, name='mg_x_input')

    mg_x_shape = K.int_shape(mg_x_input)[1:]
    mg_x_flatten = Flatten(name='mg_x_flatten')(mg_x_input)

    mg_x_encoder = Model(mg_x_input, mg_x_flatten, name='mg_x_encoder')

    ### Motiongram y Encoder ###

    # Here the input is just passed through in order to keep the number of parameters and therefore the inference time low
    # On more powerful computers extra layers can be added here in order to increase the performance of the model

    mg_y_input = Input(shape=input_3_size, name='mg_y_input')

    mg_y_shape = K.int_shape(mg_y_input)[1:]
    mg_y_flatten = Flatten(name='mg_y_flatten')(mg_y_input)

    mg_y_encoder = Model(mg_y_input, mg_y_flatten, name='mg_y_encoder')

    ### Concatenation Model ###

    # Concatenates the outputs of the audio and motiongram encoders for passing to the latent space encoder

    conc_input_1 = Input(shape=input_1_size, name='conc_input_1')
    conc_input_2 = Input(shape=aspect_x*segment_frames, name='conc_input_2')
    conc_input_3 = Input(shape=aspect_y*segment_frames, name='conc_input_3')

    conc = Concatenate()([conc_input_1, conc_input_2, conc_input_3])

    conc_model = Model(inputs=[conc_input_1, conc_input_2, conc_input_3], outputs = conc, name='conc_model')

    ### Latent Space ###

    latent_input = Input(shape=conc_model.layers[-1].output_shape[1], name='latent_input')

    mean = Dense(units=latent_space_size, name='mean')(latent_input)
    log_variance = Dense(units=latent_space_size, name='log_variance')(latent_input)

    z = Lambda(sampling_function, name='z')([mean, log_variance])

    encoder_normalise = BatchNormalization(name='encoder_normalise')(z) # batch normalisation layer included to combat exploding gradients
    encoder_output = Activation('tanh', name='encoder_activation')(encoder_normalise)

    latent_group = Model(inputs=latent_input, outputs=encoder_output, name='latent_group')

    ### Entire Encoder Model ###

    encoder_input_1 = Input(shape=input_1_size, name='encoder_input_1')
    encoder_input_2 = Input(shape=input_2_size, name='encoder_input_2')
    encoder_input_3 = Input(shape=input_3_size, name='encoder_input_3')
    audio_encoder_output = audio_encoder(encoder_input_1)
    mg_x_encoder_output = mg_x_encoder(encoder_input_2)
    mg_y_encoder_output = mg_y_encoder(encoder_input_3)
    conc_output = conc_model([audio_encoder_output, mg_x_encoder_output, mg_y_encoder_output])
    latent_group_output = latent_group(conc_output)
    encoder = Model(inputs=[encoder_input_1, encoder_input_2, encoder_input_3], outputs=latent_group_output, name='encoder')

    ### Decoder ###

    # This is structured similarly to the encoder model, with a seperate model for the audio, mg x, and mg y
    # decoders. These are then put together as an entire decoder model. L2 regularisers are employed in the dense
    # layers in order to combat exploding gradients, that can happen when the reconstruction loss weighting is
    # set to a high value.

    ### Audio Decoder ###

    audio_decoder_input = Input(latent_space_size, name='audio_decoder_output')

    audio_decoder_dense = Dense(input_1_size, name='audio_decoder_dense', activity_regularizer=regularizers.L2(L2_regulariser))(audio_decoder_input)
    audio_activation_decoder = Activation('tanh', name='audio_activation_decoder')(audio_decoder_dense)

    audio_decoder = Model(audio_decoder_input, audio_activation_decoder, name='audio_decoder')

    ### Motiongram x Decoder ###

    mg_x_decoder_input = Input(shape=(latent_space_size), name='mg_x_decoder_input')

    mg_x_decoder_dense_1 = Dense(units=np.prod(mg_x_shape), name='mg_x_decoder_dense_1', activity_regularizer=regularizers.L2(L2_regulariser))(mg_x_decoder_input)
    mg_x_decoder_reshape = Reshape(mg_x_shape)(mg_x_decoder_dense_1)

    mg_x_decoder = Model(mg_x_decoder_input, mg_x_decoder_reshape, name='mg_x_decoder')

    ### Motiongram y Decoder ###

    mg_y_decoder_input = Input(shape=(latent_space_size), name='mg_y_decoder_input')

    mg_y_decoder_dense_1 = Dense(units=np.prod(mg_y_shape), name='mg_y_decoder_dense_1', activity_regularizer=regularizers.L2(L2_regulariser))(mg_y_decoder_input)
    mg_y_decoder_reshape = Reshape(mg_y_shape)(mg_y_decoder_dense_1)

    mg_y_decoder = Model(mg_y_decoder_input, mg_y_decoder_reshape, name='mg_y_decoder')


    ### Entire Decoder Model ###

    decoder_input = Input(shape=(latent_space_size), name='decoder_input')
    audio_output = audio_decoder(decoder_input)
    mg_x_output = mg_x_decoder(decoder_input)
    mg_y_output = mg_y_decoder(decoder_input)  
    decoder = Model(inputs= decoder_input, outputs=[audio_output, mg_x_output, mg_y_output], name='decoder_model')

    ### Variational Autoencoder Model ###

    vae_input_1 = Input(shape=input_1_size, name='VAE_input_1')
    vae_input_2 = Input(shape=input_2_size, name='VAE_input_2')
    vae_input_3 = Input(shape=input_3_size, name='VAE_input_3')
    vae_encoder_output = encoder([vae_input_1, vae_input_2, vae_input_3])
    vae_decoder_output = decoder(vae_encoder_output)
    vae = Model(inputs=[vae_input_1, vae_input_2, vae_input_3], outputs=vae_decoder_output, name='VAE')

    ### Compile and Train the VAE ###

    vae.compile(optimizer = Adam(learning_rate = learning_rate), loss = loss_function(mean, log_variance))
    vae.fit(x=[audio, mgs_x, mgs_y], y=[audio, mgs_x, mgs_y], epochs=epochs, batch_size=batch_size, shuffle=True, validation_split=0.1)

    # Calculate R2 and MSE

    predict = vae.predict((audio, mgs_x, mgs_y))

    print('Coefficient of determination (r2 score) Audio: %.4f'% r2_score(audio, predict[0]))
    print('Mean Squared Error Audio: %.4f'% mean_squared_error(audio, predict[0]))
    print('Coefficient of determination (r2 score) Motiongram X: %.4f'% r2_score(mgs_x.flatten(), predict[1].flatten()))
    print('Mean Squared Error Motiongram X: %.4f'% mean_squared_error(mgs_x.flatten(), predict[1].flatten()))
    print('Coefficient of determination (r2 score) Motiongram Y: %.4f'% r2_score(mgs_y.flatten(), predict[2].flatten()))
    print('Mean Squared Error Motiongram Y: %.4f'% mean_squared_error(mgs_y.flatten(), predict[2].flatten()))

    ### Inference Model ###

    # Now the model for inference is created for use in the performance system.
    # This is done by taking the mg x and mg y encoder, and the audio decoder from the trained model
    # and attaching them to a replication of the latent group layers from the training model.
    # The weights are taken from the training model, and trimmed so that the weights between
    # the audio encoder neurons and the latent group are trimmed, to match the layer sizes for the inference model. 

    # set the input sizes

    predict_input_1 = Input(shape=input_2_size, name='predict_input_1')
    predict_input_2 = Input(shape=input_3_size, name='predict_input_2')

    # pass the inputs to the trained mg x and y encoder models

    predict_mg_x_encoder_output = mg_x_encoder(predict_input_1)
    predict_mg_y_encoder_output = mg_y_encoder(predict_input_2)

    # concatenate the mgx and mg y outputs

    predict_conc = Concatenate(name='predict_conc')([predict_mg_x_encoder_output, predict_mg_y_encoder_output])
    predict_mean = Dense(units=latent_space_size, name='predict_mean')(predict_conc)

    # replicate the latent group submodel from the training model
    predict_log_variance = Dense(units=latent_space_size, name='predict_log_variance')(predict_conc)
    predict_z = Lambda(sampling_function, name='predict_z')([predict_mean, predict_log_variance])
    predict_normalise = BatchNormalization(name='encoder_normalise')(predict_z)
    predict_activation = Activation('tanh', name='encoder_activation')(predict_normalise)

    # pass to the trained audio decoder

    predict_decoder = audio_decoder(predict_activation)

    # create the inference model

    predictor = Model(inputs=[predict_input_1,predict_input_2] , outputs=predict_decoder, name='predictor')

    # Get the weights from the latent group layers in the training model

    latent_mean_weights = latent_group.layers[1].get_weights()
    latent_var_weights = latent_group.layers[2].get_weights()
    latent_norm_weights = latent_group.layers[4].get_weights()

    # Remove the weights connecting the audio encoder to the latent group

    predict_mean_weights = [latent_mean_weights[0][input_1_size:], latent_mean_weights[1]]
    predict_var_weights = [latent_var_weights[0][input_1_size:], latent_var_weights[1]]

    # Set the remaining weights in the inference model

    predictor.layers[5].set_weights(predict_mean_weights)
    predictor.layers[6].set_weights(predict_var_weights)
    predictor.layers[8].set_weights(latent_norm_weights)

    # Save the weights for the prediction model to be used in the performance system

    predictor.save_weights('./Temp/Predictor_Weights.h5')