import numpy as np
import cv2
from Utils import motion
import threading
import time
import pyaudio
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Dense, Activation, Lambda, Concatenate, BatchNormalization, Flatten
import tensorflow.keras.backend as K

# Function that runs the prediction on a separate thread so as not to interrupt programme flow

def predictor(array_in):

    # convert to array

    motion_array = np.array(array_in)

    # get motion data

    mg_x, mg_y, _, _, _ = motion(motion_array)

    # format for model

    mg_x = mg_x.reshape(mg_x.shape[0], mg_x.shape[1], 1)
    mg_y = mg_y.reshape(mg_y.shape[0], mg_y.shape[1], 1)

    mg_x = (mg_x/128) - 1
    mg_y = (mg_y/128) - 1

    mg_x = mg_x[None, :, :, :]
    mg_y = mg_y[None, :, :, :]

    # predict

    output = model.predict((mg_x, mg_y))

    output = np.array(output).flatten()

    # place in the audio_data array which is streamed by Pyaudio
    
    global audio_data
    audio_data = output

# Sampling function for model

def sampling_function(args):
    
    mean, log_variance = args

    epsilon = K.random_normal(shape=(K.shape(mean)))

    z = mean + K.exp(log_variance/2) * epsilon

    return z

# Model function to be called to create model for inference

def model(input_1_size, input_2_size, input_3_size, latent_space_size):

    ### This is a recreation of the prediction model from the the training model script ###
    # This is called and then the saved weights are loaded.
    # This is a rather inelegant method to do this, although it was the best way that I could find before running out of time.
    # Saving the entire model (including architecture) in the training model script and then attempting to load it
    # caused an error due to the sampling function lambda layer.
    # Creating a class and calling that in the training script and then calling it here and loading weights 
    # caused problems with training with the custom loss function, and would have required implementing the model
    # using the graph API.
    # So although this is rather cumbersome, as a modification in the model trained requires the same modifications
    # to be made here, this is currently the best solution that could be found.

    ### Motiongram x Encoder ###

    mg_x_input = Input(shape=input_2_size, name='mg_x_input')

    mg_x_flatten = Flatten(name='mg_x_flatten')(mg_x_input)

    mg_x_encoder = Model(mg_x_input, mg_x_flatten, name='mg_x_encoder')

    ### Motiongram y Encoder ###

    mg_y_input = Input(shape=input_3_size, name='mg_y_input')

    mg_y_flatten = Flatten(name='mg_y_flatten')(mg_y_input)

    mg_y_encoder = Model(mg_y_input, mg_y_flatten, name='mg_y_encoder')


    ### Audio Decoder ###

    audio_decoder_input = Input(latent_space_size, name='audio_decoder_output')

    audio_decoder_dense = Dense(input_1_size, name='audio_decoder_dense')(audio_decoder_input)
    audio_activation_decoder = Activation('tanh', name='audio_activation_decoder')(audio_decoder_dense)

    audio_decoder = Model(audio_decoder_input, audio_activation_decoder, name='audio_decoder')

    ### Create Inference Model ###

    # Inputs

    predict_input_1 = Input(shape=input_2_size, name="predict_input_1")
    predict_input_2 = Input(shape=input_3_size, name="predict_input_2")

    # Pass inputs to mgx and y encoder
    predict_mg_x_encoder_output = mg_x_encoder(predict_input_1)
    predict_mg_y_encoder_output = mg_y_encoder(predict_input_2)

    # Concatenate mg x and y

    predict_conc = Concatenate(name='predict_conc')([predict_mg_x_encoder_output, predict_mg_y_encoder_output])

    # Create latent space

    predict_mean = Dense(units=latent_space_size, name="predict_mean")(predict_conc)
    predict_log_variance = Dense(units=latent_space_size, name="predict_log_variance")(predict_conc)
    predict_z = Lambda(sampling_function, name="predict_z")([predict_mean, predict_log_variance])
    predict_normalise = BatchNormalization(name='encoder_normalise')(predict_z)
    predict_activation = Activation('tanh', name='encoder_activation')(predict_normalise)

    # Pass to audio decoder

    predict_decoder = audio_decoder(predict_activation)

    # Create model

    predictor = Model(inputs=[predict_input_1,predict_input_2] , outputs=predict_decoder, name='predictor')

    return predictor

# Callback function for Pyaudio stream, reads from audio_data array
# This is used as the time needed for inference varies. This means that if the resulting infered data
# is simply written to the stream there are either audio dropouts if the inference time is lower than the
# length of the inferred segments or there is a buffer overflow if the inference time is longer than
# the length of the infered segment.
# The drawback to this method though is that the Pyaudio stream simply reads whatever is in the audio_data array
# when it has finished playing the last segment. This means that some inferred segments may be skipped if the inference time
# is longer than the segment length, or some may be played twice if it is shorter.

def callback(in_data, frame_count, time_info, status):

    global audio_data
    data = audio_data
    return (data, pyaudio.paContinue)



if __name__ == "__main__":

    # set variables

    # video

    aspect_ratio = 4/3 # set the aspect ratio
    fps = 60 # set the frames per second - This needs to be the same as the camera frame rate
    input_camera = 1 # set the input camera device, 0 = internal camera, 1 = external camera
    aspect_x = 64 # aspect ratio for x
    aspect_y = 48 #  aspect ratio for y - both x and y are temps, will default to nearest and these are taken later in script

    # audio

    output_audio_device = 4 # sets output device to play audio
                            # to find correct device index run the device index script
    fs = 8040 # set sampling frequency - should be set the same as was returned in capture script
    sample_format = pyaudio.paFloat32 # set to 32-bit float
    channels = 1 # set to mono audio
    chunk = int(fs/fps) # find the number of audio samples per frame
    segment_length = 0.5 # length of sample segment in seconds
    segment_frames = int(fps*segment_length) # set the number of video frames per segment
    segment_chunk = int(chunk*segment_frames) # set the number of audio samples per segment

    # model

    input_1_size = segment_chunk # set audio input size
    input_2_size = segment_frames, aspect_x, 1 # set motiongram x input size
    input_3_size = aspect_y, segment_frames, 1 # set motiongram y input size
    latent_space_size = 512 # set latent space size

    # load model

    model = model(input_1_size, input_2_size, input_3_size, latent_space_size)
    model.load_weights('./Temp/Predictor_Weights.h5')

    # initialise video (change to 0 to use internal camera)

    vid = cv2.VideoCapture(input_camera)

    # set fps

    vid.set(cv2.CAP_PROP_FPS, fps)
    
    # set resolution (will default to nearest)

    vid.set(cv2.CAP_PROP_FRAME_WIDTH, 1280)
    vid.set(cv2.CAP_PROP_FRAME_HEIGHT, 720)

    # get actual resolution

    width = vid.get(3)
    height = vid.get(4)

    # set new width

    new_width = height*aspect_ratio

    # find difference in widths

    difference = width - new_width

    # find array values to trim

    left = difference/2
    right = width - left

    # create array to write to to send to model

    write_array = np.empty((segment_frames, aspect_y, aspect_x))

    # set counter

    counter = 0

    # create Pyaudio stream

    p = pyaudio.PyAudio()
    stream = p.open(format=sample_format,
                channels=channels,
                rate=fs,
                frames_per_buffer=segment_chunk,
                output=True,
                output_device_index=output_audio_device,
                stream_callback=callback
                )

    audio_data  = np.zeros(segment_chunk)

    while(True):

        # video

        # read the frame from camera

        ret, frame = vid.read()

        # trim the frame to 4:3

        frame = frame[0:, int(left):int(right)]

        # lower resolution

        frame = cv2.resize(frame, (aspect_x, aspect_y))

        # flip the image

        frame = cv2.flip(frame, 1)

        # convert to greyscale

        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

        # create resizable window for display

        cv2.namedWindow('frame', cv2.WINDOW_NORMAL)

        # resize the window

        cv2.resizeWindow('frame', int(new_width), int(height))

        # display frame

        cv2.imshow('frame', frame)

        # append frame to list

        write_array[counter%segment_frames] = frame

        # using thread to predict and play so as not to break programme flow
            
        if counter%segment_frames == segment_frames-1:

            thread_start = time.time()

            t1 = threading.Thread(target=predictor, args=(write_array,))

            t1.start()
            

          # break if q key pressed

          
        if cv2.waitKey(1) & 0xFF == ord('q'):
            stream.close()
            break
                
        counter += 1
        
            
    vid.release()
    cv2.destroyAllWindows()

